#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import html

try:
    from cadquery_massembly import MAssembly

    HAS_MASSEMBLY = True
except:
    HAS_MASSEMBLY = False
import numpy as np
from cadquery.occ_impl.shapes import Face, Edge, Wire
from cadquery import Workplane, Shape, Compound, Vector, Vertex, Location, Assembly as CqAssembly

from jupyter_cadquery.cad_objects import (
    _PartGroup,
    _Part,
    _Edges,
    _Faces,
    _Vertices,
    _show,
)

from jupyter_cadquery.cad_display import get_default
from .cqparts import is_cqparts, convert_cqparts
from ..utils import Color
from ..ocp_utils import get_rgb
from ..defaults import get_default


class Part(_Part):
    def __init__(self, shape, name="Part", color=None, show_faces=True, show_edges=True):
        if color is None:
            color = get_default("default_color")
        super().__init__(_to_occ(shape), name, color, show_faces, show_edges)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=False, axes=False):
        return show(self, grid=grid, axes=axes)


class Faces(_Faces):
    def __init__(self, faces, name="Faces", color=None, show_faces=True, show_edges=True):
        super().__init__(_to_occ(faces.combine()), name, color, show_faces, show_edges)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=False, axes=False):
        return show(self, grid=grid, axes=axes)


class Edges(_Edges):
    def __init__(self, edges, name="Edges", color=None):
        super().__init__(_to_occ(edges), name, color)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=False, axes=False):
        return show(self, grid=grid, axes=axes)


class Vertices(_Vertices):
    def __init__(self, vertices, name="Vertices", color=None):
        super().__init__(_to_occ(vertices), name, color)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=False, axes=False):
        return show(self, grid=grid, axes=axes)


class PartGroup(_PartGroup):
    def to_assembly(self):
        return self

    def show(self, grid=False, axes=False):
        return show(self, grid=grid, axes=axes)

    def add(self, cad_obj):
        self.objects.append(cad_obj)

    def add_list(self, cad_objs):
        self.objects += cad_objs


class Assembly(PartGroup):
    def __init__(self, *args, **kwargs):
        import warnings

        super().__init__(*args, **kwargs)
        warnings.warn(
            "Class 'Assembly' is deprecated (too many assemblies ...). Please use class 'PartGroup' instead",
            RuntimeWarning,
        )


def _to_occ(cad_obj):
    # special case Wire, must be handled before Workplane
    if _is_wirelist(cad_obj):
        all_edges = []
        for edges in cad_obj.objects:
            all_edges += edges.Edges()
        return [edge.wrapped for edge in all_edges]

    elif isinstance(cad_obj, Workplane):
        return [obj.wrapped for obj in cad_obj.objects]

    elif isinstance(cad_obj, Shape):
        return [cad_obj.wrapped]

    else:
        raise NotImplementedError(type(cad_obj))


def _parent(cad_obj, obj_id):
    if cad_obj.parent is not None:
        if isinstance(cad_obj.parent.val(), Vector):
            return _from_vectorlist(
                cad_obj.parent,
                obj_id,
                name="Parent",
                color=Color((0.8, 0.8, 0.8)),
                show_parents=False,
            )
        elif isinstance(cad_obj.parent.val(), Vertex):
            return _from_vertexlist(
                cad_obj.parent,
                obj_id,
                name="Parent",
                color=Color((0.8, 0.8, 0.8)),
                show_parents=False,
            )
        elif isinstance(cad_obj.parent.val(), Edge):
            return _from_edgelist(
                cad_obj.parent,
                obj_id,
                name="Parent",
                color=Color((0.8, 0.8, 0.8)),
                show_parents=False,
            )
        elif isinstance(cad_obj.parent.val(), Wire):
            return [_from_wirelist(cad_obj.parent, obj_id, name="Parent", color=Color((0.8, 0.8, 0.8)))]
        else:
            return [
                Part(
                    cad_obj.parent,
                    "Parent_%d" % obj_id,
                    show_edges=True,
                    show_faces=False,
                )
            ]
    else:
        return []


def _from_facelist(cad_obj, obj_id, name="Faces", show_parents=True):
    result = [Faces(cad_obj, "%s_%d" % (name, obj_id), color=Color((0.8, 0.0, 0.8)))]
    if show_parents:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_edgelist(cad_obj, obj_id, name="Edges", color=None, show_parents=True):
    result = [Edges(cad_obj, "%s_%d" % (name, obj_id), color=Color(color or (1.0, 0.0, 1.0)))]
    if show_parents:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_vector(vec, obj_id, name="Vector"):
    tmp = Workplane()
    obj = tmp.newObject([vec])
    return _from_vectorlist(obj, obj_id, name)


def _from_vectorlist(cad_obj, obj_id, name="Vertices", color=None, show_parents=True):
    if cad_obj.vals():
        vectors = cad_obj.vals()
    else:
        vectors = [cad_obj.val()]
    obj = cad_obj.newObject([Vertex.makeVertex(v.x, v.y, v.z) for v in vectors])
    result = [Vertices(obj, "%s_%d" % (name, obj_id), color=Color(color or (1.0, 0.0, 1.0)))]
    if show_parents:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_vertexlist(cad_obj, obj_id, name="Vertices", color=None, show_parents=True):
    result = [Vertices(cad_obj, "%s_%d" % (name, obj_id), color=Color(color or (1.0, 0.0, 1.0)))]
    if show_parents:
        result = _parent(cad_obj, obj_id) + result
    return result


def _from_wirelist(cad_obj, obj_id, name="Edges", color=None):
    return Edges(cad_obj, "%s_%d" % (name, obj_id), color=Color(color or (1.0, 0.0, 1.0)))


def to_edge(mate, loc=None, scale=1) -> Workplane:
    w = Workplane()
    for d in (mate.x_dir, mate.y_dir, mate.z_dir):
        edge = Edge.makeLine(mate.origin, mate.origin + d * scale)
        w.objects.append(edge if loc is None else edge.moved(loc))

    return w


def from_assembly(cad_obj, top, loc=None, render_mates=False, mate_scale=1, default_color=None):
    loc = Location()
    render_loc = cad_obj.loc

    if cad_obj.color is None:
        if default_color is None:
            color = Color(get_default("default_color"))
        else:
            color = Color(default_color)
    else:
        color = Color(get_rgb(cad_obj.color))

    parent = [
        Part(
            Workplane(shape),
            "%s_%d" % (cad_obj.name, i),
            color=color,
        )
        for i, shape in enumerate(cad_obj.shapes)
    ]

    if render_mates and cad_obj.mates is not None:
        RGB = (Color((255, 0, 0)), Color((0, 128, 0)), Color((0, 0, 255)))
        parent.append(
            PartGroup(
                [
                    Edges(to_edge(mate_def.mate, scale=mate_scale), name=name, color=RGB)
                    for name, mate_def in top.mates.items()
                    if mate_def.assembly == cad_obj
                ],
                name="mates",
                loc=Location(),  # mates inherit the parent location, so actually add a no-op
            )
        )

    children = [from_assembly(c, top, loc, render_mates, mate_scale) for c in cad_obj.children]
    return PartGroup(parent + children, cad_obj.name, loc=render_loc)


def _from_workplane(cad_obj, obj_id, name="Part", default_color=None):
    return Part(cad_obj, "%s_%d" % (name, obj_id), color=Color(default_color))


def _is_facelist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Face) for obj in cad_obj.objects])
    )


def _is_vertexlist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Vertex) for obj in cad_obj.objects])
    )


def _is_edgelist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Edge) for obj in cad_obj.objects])
    )


def _is_wirelist(cad_obj):
    return (
        hasattr(cad_obj, "objects")
        and cad_obj.objects != []
        and all([isinstance(obj, Wire) for obj in cad_obj.objects])
    )


def to_assembly(*cad_objs, render_mates=None, mate_scale=1, default_color=None):
    default_color = get_default("default_color") if default_color is None else default_color
    assembly = PartGroup([], "Group")
    obj_id = 0
    for cad_obj in cad_objs:
        if isinstance(cad_obj, (PartGroup, Part, Faces, Edges, Vertices)):
            assembly.add(cad_obj)

        elif HAS_MASSEMBLY and isinstance(cad_obj, MAssembly):
            assembly.add(
                from_assembly(
                    cad_obj, cad_obj, render_mates=render_mates, mate_scale=mate_scale, default_color=default_color
                )
            )

        elif isinstance(cad_obj, CqAssembly):
            assembly.add(from_assembly(cad_obj, cad_obj, default_color=default_color))

        elif isinstance(cad_obj, Edge):
            assembly.add_list(_from_edgelist(Workplane(cad_obj), obj_id))

        elif isinstance(cad_obj, Face):
            assembly.add_list(_from_facelist(Workplane(cad_obj), obj_id))

        elif isinstance(cad_obj, Wire):
            assembly.add(_from_wirelist(Workplane(cad_obj), obj_id))

        elif isinstance(cad_obj, Vertex):
            assembly.add_list(_from_vertexlist(Workplane(cad_obj), obj_id))

        elif is_cqparts(cad_obj):
            assembly = convert_cqparts(cad_obj)

        elif _is_facelist(cad_obj):
            assembly.add_list(_from_facelist(cad_obj, obj_id))

        elif _is_edgelist(cad_obj):
            assembly.add_list(_from_edgelist(cad_obj, obj_id))

        elif _is_wirelist(cad_obj):
            assembly.add(_from_wirelist(cad_obj, obj_id))

        elif _is_vertexlist(cad_obj):
            assembly.add_list(_from_vertexlist(cad_obj, obj_id))

        elif isinstance(cad_obj, Vector):
            assembly.add_list(_from_vector(cad_obj, obj_id))

        elif isinstance(cad_obj, (Shape, Compound)):
            assembly.add(_from_workplane(Workplane(cad_obj), obj_id, default_color=default_color))

        elif isinstance(cad_obj.val(), Vector):
            assembly.add_list(_from_vectorlist(cad_obj, obj_id))

        elif isinstance(cad_obj, Workplane):
            assembly.add(_from_workplane(cad_obj, obj_id, default_color=default_color))

        else:
            raise NotImplementedError("Type:", cad_obj)

        obj_id += 1
    return assembly


def show(*cad_objs, render_mates=None, mate_scale=None, **kwargs):
    """Show CAD objects in Jupyter

    Valid keywords:
    - height:            Height of the CAD view (default=600)
    - tree_width:        Width of navigation tree part of the view (default=250)
    - cad_width:         Width of CAD view part of the view (default=800)
    - bb_factor:         Scale bounding box to ensure compete rendering (default=1.5)
    - default_color:     Default mesh color (default=(232, 176, 36))
    - default_edgecolor: Default mesh color (default=(128, 128, 128))
    - render_edges:      Render edges  (default=True)
    - render_normals:    Render normals (default=False)
    - render_mates:      Render mates (for MAssemblies)
    - mate_scale:        Scale of rendered mates (for MAssemblies)
    - quality:           Linear deflection for tessellation (default=None)
                         If None, uses bounding box as in (xlen + ylen + zlen) / 300 * deviation)
    - deviation:         Deviation from default for linear deflection value ((default=0.1)
    - angular_tolerance: Angular deflection in radians for tessellation (default=0.2)
    - edge_accuracy:     Presicion of edge discretizaion (default=None)
                         If None, uses: quality / 100
    - optimal_bb:        Use optimal bounding box (default=False)
    - axes:              Show axes (default=False)
    - axes0:             Show axes at (0,0,0) (default=False)
    - grid:              Show grid (default=False)
    - ticks:             Hint for the number of ticks in both directions (default=10)
    - ortho:             Use orthographic projections (default=True)
    - transparent:       Show objects transparent (default=False)
    - ambient_intensity  Intensity of ambient ligth (default=1.0)
    - direct_intensity   Intensity of direct lights (default=0.12)
    - position:          Relative camera position that will be scaled (default=(1, 1, 1))
    - rotation:          z, y and y rotation angles to apply to position vector (default=(0, 0, 0))
    - zoom:              Zoom factor of view (default=2.5)
    - reset_camera:      Reset camera position, rotation and zoom to default (default=True)
    - mac_scrollbar:     Prettify scrollbars (default=True)
    - display:           Select display: "sidecar", "cell", "html"
    - tools:             Show the viewer tools like the object tree
    - timeit:            Show rendering times, levels = False, 0,1,2,3,4,5 (default=False)

    For example isometric projection can be achieved in two ways:
    - position = (1, 1, 1)
    - position = (0, 0, 1) and rotation = (45, 35.264389682, 0)
    """
    render_mates = render_mates or get_default("render_mates")
    mate_scale = mate_scale or get_default("mate_scale")
    default_color = kwargs.get("default_color") or get_default("default_color")

    assembly = to_assembly(*cad_objs, render_mates=render_mates, mate_scale=mate_scale, default_color=default_color)

    if assembly is None:
        raise ValueError("%s cannot be viewed" % cad_objs)

    if len(assembly.objects) == 1 and isinstance(assembly.objects[0], PartGroup):
        # omit leading "PartGroup" group
        return _show(assembly.objects[0], **kwargs)
    else:
        return _show(assembly, **kwargs)


def auto_show():
    PartGroup._ipython_display_ = lambda self: self.show()
    Part._ipython_display_ = lambda self: self.show()
    Faces._ipython_display_ = lambda self: self.show(grid=False, axes=False)
    Edges._ipython_display_ = lambda self: self.show(grid=False, axes=False)
    Vertices._ipython_display_ = lambda self: self.show(grid=False, axes=False)

    print("Overwriting auto display for cadquery Workplane and Shape")

    import cadquery as cq

    try:
        del cq.Workplane._repr_html_
        del cq.Shape._repr_html_
    except:
        pass
    cq.Workplane._ipython_display_ = lambda cad_obj: show(cad_obj)
    cq.Shape._ipython_display_ = lambda cad_obj: show(cad_obj)
    cq.Assembly._ipython_display_ = lambda cad_obj: show(cad_obj)


# Some further cq.Assembly methods


def show_constraints(assy, qs):
    colors = [
        "#e41a1c",
        "#377eb8",
        "#4daf4a",
        "#984ea3",
        "#ff7f00",
        "#ffff33",
        "#a65628",
        "#f781bf",
        "#999999",
        "#8dd3c7",
        "#ffffb3",
        "#bebada",
        "#fb8072",
        "#80b1d3",
        "#fdb462",
        "#b3de69",
        "#fccde5",
        "#d9d9d9",
    ]

    constraints = []
    objects = []
    cache = {}

    for i, q1q2 in enumerate(qs):
        parts = []

        kind = q1q2[-1]

        if len(q1q2) == 3:
            q1q2 = ((q1q2[0].split("@")[0], q1q2[0]), (q1q2[1].split("@")[0], q1q2[1]))
        else:
            q1q2 = (q1q2[0:2], q1q2[2:4])

        for q in q1q2:
            name, shape = q
            if name in cache:
                obj = cache[name]["obj"]
                loc = cache[name]["loc"]
            else:
                obj = assy.objects[name].obj
                loc = assy.objects[name].loc

                parent = assy.objects[name].parent
                while parent is not None:
                    loc = parent.loc * loc
                    parent = parent.parent

                cache[name] = {"obj": obj, "loc": loc, "shape": shape}

                objects.append(Part(Workplane(obj.val().located(loc)), name=name, show_faces=False))

            label = str(shape)
            if isinstance(shape, str):
                shape = assy._query(shape)[1]

            parts.append(
                Faces(
                    Workplane(Workplane(shape).val().located(loc)),
                    name=html.escape(label),
                    color=colors[i % len(colors)],
                )
            )
        constraints.append(PartGroup(parts, "%s_%d" % (kind, i)))

    show(PartGroup([PartGroup(objects, "objects")] + constraints), axes=True, axes0=True)


def show_accuracy(assy, cs):
    def relocate(name, shape):
        a = assy.objects[name]
        loc = a.loc

        parent = a.parent
        while parent is not None:
            loc = parent.loc * loc
            parent = parent.parent

        if isinstance(shape, str):
            shape = assy._query(shape)[1]
        return Workplane(Workplane(shape).val().located(loc))

    def center(face):
        c = face.Center()
        return np.array((c.x, c.y, c.z))

    def normal(face):
        n = face.normalAt()
        return np.array((n.x, n.y, n.z))

    def print_metric(results):
        l = max([len(r[1]) for r in results])
        h = ("Constraint", "Normal-Dist", "Normal-Angle", "Point-Dist")
        print(f"{h[0]:{l+7}s} {h[1]:12s}  {h[2]:12s}  {h[3]:12s}")
        print("-" * (l + 46))
        for kind, label, nrm_dist, nrm_angle, pnt_dist in results:
            metric = f"{kind:5s} {label:{l}s} "
            metric += " " * 27 if nrm_dist is None else f"{nrm_dist:12.9f}  {nrm_angle:12.8}°"
            metric += " " * 13 if pnt_dist is None else f"{pnt_dist:12.9f}"
            print(metric)

    results = []
    for q1q2 in cs:
        kind = q1q2[-1]

        if len(q1q2) == 3:
            n_q1q2 = ((q1q2[0].split("@")[0], q1q2[0]), (q1q2[1].split("@")[0], q1q2[1]))
            label = "%s - %s" % q1q2[:2]
        else:
            n_q1q2 = (q1q2[0:2], q1q2[2:4])
            label = "%s<%s> - %s<%s>" % (q1q2[0], q1q2[1].__class__.__name__, q1q2[2], q1q2[3].__class__.__name__)

        shape1 = relocate(*n_q1q2[0])
        shape2 = relocate(*n_q1q2[1])

        pnt_dist = None
        nrm_dist = None
        nrm_angle = None

        if kind in ["Point", "Plane"]:
            c1, c2 = center(shape1.val()), center(shape2.val())
            pnt_dist = np.linalg.norm(c1 - c2)
        if kind in ["Axis", "Plane"]:
            n1, n2 = normal(shape1.val()), normal(shape2.val())
            nrm_dist = np.linalg.norm(n1 + n2)  # distance between n1 and -n2 since n1 and n2 point opposite
            c = np.dot(n1, -n2) / np.linalg.norm(n1) / np.linalg.norm(n2)
            nrm_angle = np.arccos(np.clip(c, -1, 1)) / np.pi * 180

        results.append((kind, label, nrm_dist, nrm_angle, pnt_dist))

    print_metric(results)



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .cad_objects import (
    Assembly,
    PartGroup,
    Part,
    Faces,
    Edges,
    Vertices,
    show,
    auto_show,
    show_accuracy,
    show_constraints,
)
from .replay import replay, enable_replay, disable_replay, reset_replay

try:
    from IPython import get_ipython

    shell_name = get_ipython().__class__.__name__
    if shell_name == "ZMQInteractiveShell":
        auto_show()
except:
    ...



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from pythreejs import (
        BufferAttribute,
        BufferGeometry,
        Mesh,
        LineSegmentsGeometry,
        LineMaterial,
        LineSegments2,
        Points,
        PointsMaterial,
        Group,
    )

from .cad_helpers import CustomMaterial

from .utils import (
    Color,
    tree_find_single_selector,
    Timer,
)


def material(color, transparent=False, opacity=1.0):
    material = CustomMaterial("standard")
    material.color = color
    material.clipping = True
    material.side = "DoubleSide"
    material.alpha = 0.7
    material.polygonOffset = True
    material.polygonOffsetFactor = 1
    material.polygonOffsetUnits = 1
    material.transparent = transparent
    material.opacity = opacity
    material.update("metalness", 0.3)
    material.update("roughness", 0.8)
    return material


class IndexedGroup(Group):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ind = None

    def dump(self, ind=""):
        print(ind, self.name, self.ind)
        for c in self.children:
            if isinstance(c, IndexedGroup):
                c.dump(ind + "    ")
            else:
                print(ind + "  ", c)

    def find_group(self, query):
        return tree_find_single_selector(self, query)


class IndexedMesh(Mesh):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ind = None

    def __repr__(self):
        return (
            f"IndexedMesh(name='{self.name}', ind={self.ind}, position={self.position}, quaternion={self.quaternion})"
        )


class IndexedPoints(Points):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ind = None

    def __repr__(self):
        return f"IndexedPoints(name='{self.name}', ind={self.ind}, position={self.position}, quaternion={self.quaternion})"


class IndexedLineSegments2(LineSegments2):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ind = None

    def __repr__(self):
        return f"IndexedLineSegments2(name='{self.name}', ind={self.ind}, position={self.position}, quaternion={self.quaternion})"


class CadqueryRenderer(object):
    def __init__(
        self,
        default_mesh_color=None,
        default_edge_color=None,
        timeit=False,
    ):
        self.default_mesh_color = Color(default_mesh_color or (166, 166, 166))
        self.default_edge_color = Color(default_edge_color or (128, 128, 128))

        self.timeit = timeit

    def _render_shape(
        self,
        shape=None,
        edges=None,
        vertices=None,
        mesh_color=None,
        edge_color=None,
        vertex_color=None,
        edge_width=1,
        vertex_width=5,
        transparent=False,
        opacity=1.0,
    ):

        edge_list = []
        normals_list = []
        edge_lines = []
        normal_lines = []
        points = None
        shape_mesh = None

        # edge_accuracy = None

        if shape is not None:
            # Compute the tesselation and build mesh
            with Timer(self.timeit, "", "build mesh:", 5):
                edge_list, normals_list = shape["edges"]
                shape_geometry = BufferGeometry(
                    attributes={
                        "position": BufferAttribute(shape["vertices"]),
                        "index": BufferAttribute(shape["triangles"]),
                        "normal": BufferAttribute(shape["normals"]),
                    }
                )

                if mesh_color is None:
                    mesh_color = self.default_mesh_color
                shp_material = material(mesh_color, transparent=transparent, opacity=opacity)
                shape_mesh = IndexedMesh(geometry=shape_geometry, material=shp_material)

        if vertices is not None:
            if vertex_color is None:
                vertex_color = self.default_edge_color  # same as edge_color

            vertices_list = vertices
            attributes = {"position": BufferAttribute(vertices_list, normalized=False)}

            mat = PointsMaterial(color=vertex_color, sizeAttenuation=False, size=vertex_width)
            geom = BufferGeometry(attributes=attributes)
            points = IndexedPoints(geometry=geom, material=mat)

        if edges is not None:
            edge_list = edges

        if len(edge_list) > 0:
            if edge_color is None:
                edge_color = self.default_edge_color

            if isinstance(edge_color, (list, tuple)):
                if len(edge_list) != len(edge_color):
                    print("warning: color list and edge list have different length, using first color for all edges")
                    edge_color = edge_color[0]

            if isinstance(edge_color, (list, tuple)):
                lines = LineSegmentsGeometry(
                    positions=edge_list,
                    colors=[[Color(color).percentage] * 2 for color in edge_color],
                )
                mat = LineMaterial(linewidth=edge_width, vertexColors="VertexColors")
                edge_lines = [IndexedLineSegments2(lines, mat)]
            else:
                lines = LineSegmentsGeometry(positions=edge_list)
                mat = LineMaterial(
                    linewidth=edge_width, color=edge_color.web_color if isinstance(edge_color, Color) else edge_color
                )
                edge_lines = [IndexedLineSegments2(lines, mat)]

        if len(normals_list) > 0:
            lines = LineSegmentsGeometry(positions=normals_list)
            mat = LineMaterial(linewidth=2, color="#9400d3")
            normal_lines = [IndexedLineSegments2(lines, mat)]

        return shape_mesh, edge_lines, normal_lines, points

    def _render(self, shapes, current, prefix=""):

        group = IndexedGroup()
        # we need to ensure unique names to enable Threejs animation later which currently doesn't
        # support directory names
        _, _, name = shapes["name"].rpartition("/")
        group.name = name if prefix == "" else f"{prefix}\\{name}"
        group.ind = current

        if shapes["loc"] is not None:
            group.position, group.quaternion = shapes["loc"]

        # Render all shapes
        for shape in shapes["parts"]:
            if shape.get("parts") is None:
                self._mapping[shape["ind"]] = {"mesh": None, "edges": None}

                if shape["type"] == "edges":
                    options = dict(
                        edges=shape["shape"],
                        edge_color=shape["color"],
                        edge_width=3,
                    )
                elif shape["type"] == "vertices":
                    options = dict(
                        vertices=shape["shape"],
                        vertex_color=shape["color"],
                        vertex_width=6,
                    )
                else:
                    options = dict(
                        shape=shape["shape"],
                        mesh_color=shape["color"],
                    )

                with Timer(self.timeit, shape["name"], "render shape:", 4):
                    shape_mesh, edge_lines, normal_lines, points = self._render_shape(**options)

                ind = len(group.children)
                if shape_mesh is not None:
                    shape_mesh.name = shape["name"]
                    shape_mesh.ind = {"group": (*current, ind), "shape": shape["ind"]}
                    shape_mesh.visible = False
                    group.add(shape_mesh)
                    self._mapping[shape["ind"]]["mesh"] = (*current, ind)
                    ind += 1

                if edge_lines or normal_lines:
                    edge_group = IndexedGroup()
                    edge_group.name = "edges"
                    edge_group.ind = (*current, ind)
                    for j, edge in enumerate(edge_lines + normal_lines):
                        edge.name = shape["name"]
                        edge.ind = {"group": (*current, ind, j), "shape": shape["ind"]}
                        edge_group.add(edge)
                    group.add(edge_group)
                    edge_group.visible = False
                    self._mapping[shape["ind"]]["edges"] = (*current, ind)
                    ind += 1

                if points is not None:
                    points.name = shape["name"]
                    points.ind = {"group": (*current, ind), "shape": shape["ind"]}
                    group.add(points)
                    self._mapping[shape["ind"]]["mesh"] = (*current, ind)
                    ind += 1

                self.progress.update()
            else:
                ind = len(group.children)
                group.add(self._render(shape, (*current, ind), group.name))

        return group

    def render(self, shapes, progress):
        self.progress = progress
        self._mapping = {}
        rendered_objects = self._render(shapes, (), "")
        return rendered_objects, self._mapping



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import math
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from pythreejs import (
        GridHelper,
        LineSegmentsGeometry,
        LineSegments2,
        LineMaterial,
        ShaderMaterial,
        ShaderLib,
    )

import numpy as np


class Helpers(object):
    def __init__(self, bb_center):
        self.bb_center = bb_center
        self.center = (0, 0, 0)
        self.zero = True

    def _center(self, zero=True):
        self.zero = zero
        return self.center if zero else self.bb_center

    def get_position(self):
        raise NotImplementedError()

    def set_position(self, position):
        raise NotImplementedError()

    def set_visibility(self, change):
        raise NotImplementedError()

    def get_center(self):
        return self.get_position()

    def is_center(self):
        return self.zero

    def set_center(self, change):
        self.set_position(self._center(change))


class Grid(Helpers):
    def __init__(self, bb_center=None, maximum=5, ticks=10, colorCenterLine="#aaa", colorGrid="#ddd"):
        super().__init__(bb_center)
        self.maximum = maximum
        axis_start, axis_end, nice_tick = self.nice_bounds(-maximum, maximum, 2 * ticks)
        self.step = nice_tick
        self.size = axis_end - axis_start
        self.grid = GridHelper(
            self.size, int(self.size / self.step), colorCenterLine=colorCenterLine, colorGrid=colorGrid
        )
        self.set_center(True)

    # https://stackoverflow.com/questions/4947682/intelligently-calculating-chart-tick-positions
    def _nice_number(self, value, round_=False):
        exponent = math.floor(math.log(value, 10))
        fraction = value / 10 ** exponent

        if round_:
            if fraction < 1.5:
                nice_fraction = 1.0
            elif fraction < 3.0:
                nice_fraction = 2.0
            elif fraction < 7.0:
                nice_fraction = 5.0
            else:
                nice_fraction = 10.0
        else:
            if fraction <= 1:
                nice_fraction = 1.0
            elif fraction <= 2:
                nice_fraction = 2.0
            elif fraction <= 5:
                nice_fraction = 5.0
            else:
                nice_fraction = 10.0

        return nice_fraction * 10 ** exponent

    def nice_bounds(self, axis_start, axis_end, num_ticks=10):
        axis_width = axis_end - axis_start
        if axis_width == 0:
            nice_tick = 0
        else:
            nice_range = self._nice_number(axis_width)
            nice_tick = self._nice_number(nice_range / (num_ticks - 1), round_=True)
            axis_start = math.floor(axis_start / nice_tick) * nice_tick
            axis_end = math.ceil(axis_end / nice_tick) * nice_tick

        return axis_start, axis_end, nice_tick

    def get_position(self):
        return self.grid.position

    def set_position(self, position):
        self.grid.position = position

    def get_visibility(self):
        return self.grid.visible

    def set_visibility(self, change):
        self.grid.visible = change

    def set_rotation(self, rotation):
        self.grid.rotation = rotation


class Axes(Helpers):
    def __init__(self, bb_center, length=1, width=3):
        super().__init__(bb_center)

        self.axes = []
        for vector, color in zip(([length, 0, 0], [0, length, 0], [0, 0, length]), ("red", "green", "blue")):
            self.axes.append(
                LineSegments2(
                    LineSegmentsGeometry(positions=[[self.center, self._shift(self.center, vector)]]),
                    LineMaterial(linewidth=width, color=color),
                )
            )

    def _shift(self, v, offset):
        return [x + o for x, o in zip(v, offset)]

    def get_position(self):
        return self.axes[0].position

    def set_position(self, position):
        for i in range(3):
            self.axes[i].position = position

    def get_visibility(self):
        return self.axes[0].visible

    def set_visibility(self, change):
        for i in range(3):
            self.axes[i].visible = change


class CustomMaterial(ShaderMaterial):
    def __init__(self, typ):
        self.types = {
            "diffuse": "c",
            "uvTransform": "m3",
            "normalScale": "v2",
            "fogColor": "c",
            "emissive": "c",
        }

        shader = ShaderLib[typ]

        fragmentShader = """
        uniform float alpha;
        """
        frag_from = "gl_FragColor = vec4( outgoingLight, diffuseColor.a );"
        frag_to = """
            if ( gl_FrontFacing ) {
                gl_FragColor = vec4( outgoingLight, alpha * diffuseColor.a );
            } else {
                gl_FragColor = vec4( diffuseColor.r, diffuseColor.g, diffuseColor.b, alpha * diffuseColor.a );
            }"""
        fragmentShader += shader["fragmentShader"].replace(frag_from, frag_to)

        vertexShader = shader["vertexShader"]
        uniforms = shader["uniforms"]
        uniforms["alpha"] = dict(value=0.7)

        super().__init__(uniforms=uniforms, vertexShader=vertexShader, fragmentShader=fragmentShader)
        self.lights = True

    @property
    def color(self):
        return self.uniforms["diffuse"]["value"]

    @color.setter
    def color(self, value):
        self.update("diffuse", value)

    @property
    def alpha(self):
        return self.uniforms["alpha"]["value"]

    @alpha.setter
    def alpha(self, value):
        self.update("alpha", value)

    def update(self, key, value):
        uniforms = dict(**self.uniforms)
        if self.types.get(key) is None:
            uniforms[key] = {"value": value}
        else:
            uniforms[key] = {"type": self.types.get(key), "value": value}
        self.uniforms = uniforms
        self.needsUpdate = True



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from cadquery import Compound, __version__

from jupyter_cadquery.cad_display import (
    get_default,
    get_or_create_display,
    has_sidecar,
)
from jupyter_cadquery_widgets.widgets import UNSELECTED, SELECTED, EMPTY
from jupyter_cadquery.utils import Color, flatten, Timer, warn
from jupyter_cadquery.ocp_utils import bounding_box, get_point, BoundingBox, loc_to_tq
from jupyter_cadquery.tessellator import discretize_edge, tessellate, compute_quality
from jupyter_cadquery.defaults import get_default, split_args

PART_ID = 0


#
# Simple Part and PartGroup classes
#


class _CADObject(object):
    def __init__(self):
        self.color = Color(get_default("default_color"))

    def next_id(self):
        global PART_ID
        PART_ID += 1
        return PART_ID

    def to_nav_dict(self):
        raise NotImplementedError("not implemented yet")

    def to_state(self):
        raise NotImplementedError("not implemented yet")

    def collect_shapes(self, loc, quality, deviation, angular_tolerance, edge_accuracy):
        raise NotImplementedError("not implemented yet")

    def to_assembly(self):
        raise NotImplementedError("not implemented yet")

    def show(self, grid=False, axes=False):
        raise NotImplementedError("not implemented yet")


class _Part(_CADObject):
    def __init__(self, shape, name="Part", color=None, show_faces=True, show_edges=True):
        super().__init__()
        self.name = name
        self.id = self.next_id()
        self.color = Color(get_default("default_color") if color is None else color)

        self.shape = shape
        self.set_states(show_faces, show_edges)

    def set_states(self, show_faces, show_edges):
        self.state_faces = SELECTED if show_faces else UNSELECTED
        self.state_edges = SELECTED if show_edges else UNSELECTED

    def to_nav_dict(self):
        return {
            "type": "leaf",
            "name": self.name,
            "id": self.id,
            "color": self.color.web_color,
        }

    def to_state(self):
        return [self.state_faces, self.state_edges]

    def collect_shapes(
        self,
        loc,
        quality,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        render_normals,
        progress=None,
        timeit=False,
    ):

        # A first rough estimate of the bounding box.
        # Will be too large, but is sufficient for computing the quality
        with Timer(timeit, self.name, "compute quality:", 2) as t:
            bb = bounding_box(self.shape, loc=loc, optimal=False)
            quality = compute_quality(bb, deviation=deviation)
            t.info = str(bb)

        normals_len = 0 if render_normals is False else quality / deviation * 5

        with Timer(timeit, self.name, "tessellate:     ", 2) as t:
            mesh = tessellate(
                self.shape,
                quality=quality,
                angular_tolerance=angular_tolerance,
                normals_len=normals_len,
                debug=timeit,
                compute_edges=render_edges,
            )
            t.info = f"{{quality:{quality:.4f}, angular_tolerance:{angular_tolerance:.2f}}}"

        # After meshing the non optimal bounding box is much more exact
        with Timer(timeit, self.name, "bounding box:   ", 2) as t:
            bb2 = bounding_box(self.shape, loc=loc, optimal=False)
            bb2.update(bb, minimize=True)
            t.info = str(bb2)

        if progress:
            progress.update()

        color = [c.web_color for c in self.color] if isinstance(self.color, tuple) else self.color.web_color

        return {
            "id": self.id,
            "type": "shapes",
            "name": self.name,
            "shape": mesh,
            "color": color,
            "bb": bb2.to_dict(),
        }

    def compound(self):
        return self.shape[0]

    def compounds(self):
        return [self.compound()]


class _Faces(_Part):
    def __init__(self, faces, name="Faces", color=None, show_faces=True, show_edges=True):
        super().__init__(faces, name, color, show_faces, show_edges)
        self.color = Color(color or (255, 0, 255))


class _Edges(_CADObject):
    def __init__(self, edges, name="Edges", color=None):
        super().__init__()
        self.shape = edges
        self.name = name
        self.id = self.next_id()

        if color is not None:
            if isinstance(color, (list, tuple)) and isinstance(color[0], Color):
                self.color = color
            elif isinstance(color, Color):
                self.color = color
            else:
                self.color = Color(color)

    def to_nav_dict(self):
        if isinstance(self.color, (tuple, list)):
            color = [c.web_color for c in self.color]
        else:
            color = self.color.web_color
        return {
            "type": "leaf",
            "name": self.name,
            "id": self.id,
            "color": color,
        }

    def to_state(self):
        return [EMPTY, SELECTED]

    def collect_shapes(
        self,
        loc,
        quality,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        render_normals,
        progress=None,
        timeit=False,
    ):
        with Timer(timeit, self.name, "bounding box:", 2) as t:
            bb = bounding_box(self.shape, loc=loc)
            quality = compute_quality(bb, deviation=deviation)
            deflection = quality / 100 if edge_accuracy is None else edge_accuracy
            t.info = str(bb)

        with Timer(timeit, self.name, "discretize:  ", 2):
            edges = flatten([discretize_edge(edge, deflection) for edge in self.shape])

        if progress:
            progress.update()

        color = [c.web_color for c in self.color] if isinstance(self.color, tuple) else self.color.web_color

        return {
            "id": self.id,
            "type": "edges",
            "name": self.name,
            "shape": edges,
            "color": color,
            "bb": bb,
        }


class _Vertices(_CADObject):
    def __init__(self, vertices, name="Vertices", color=None):
        super().__init__()
        self.shape = vertices
        self.name = name
        self.id = self.next_id()
        self.color = Color(color or (255, 0, 255))

    def to_nav_dict(self):
        return {
            "type": "leaf",
            "name": self.name,
            "id": self.id,
            "color": self.color.web_color,
        }

    def to_state(self):
        return [SELECTED, EMPTY]

    def collect_shapes(
        self,
        loc,
        quality,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        render_normals,
        progress=None,
        timeit=False,
    ):
        bb = bounding_box(self.shape, loc=loc)

        if progress:
            progress.update()

        return {
            "id": self.id,
            "type": "vertices",
            "name": self.name,
            "shape": [get_point(vertex) for vertex in self.shape],
            "color": self.color.web_color,
            "bb": bb,
        }


class _PartGroup(_CADObject):
    def __init__(self, objects, name="Group", loc=None):
        super().__init__()
        self.objects = objects
        self.name = name
        self.loc = loc
        self.id = self.next_id()

    def to_nav_dict(self):
        return {
            "type": "node",
            "name": self.name,
            "id": self.id,
            "children": [obj.to_nav_dict() for obj in self.objects],
        }

    def collect_shapes(
        self,
        loc,
        quality,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        render_normals,
        progress=None,
        timeit=False,
    ):
        if loc is None and self.loc is None:
            combined_loc = None
        elif loc is None:
            combined_loc = self.loc
        else:
            combined_loc = loc * self.loc

        result = {"parts": [], "loc": None if self.loc is None else loc_to_tq(self.loc), "name": self.name}
        for obj in self.objects:
            result["parts"].append(
                obj.collect_shapes(
                    combined_loc,
                    quality,
                    deviation,
                    angular_tolerance,
                    edge_accuracy,
                    render_edges,
                    render_normals,
                    progress,
                    timeit,
                )
            )
        return result

    def collect_mapped_shapes(
        self,
        mapping,
        quality,
        deviation,
        angular_tolerance,
        edge_accuracy,
        render_edges,
        render_normals,
        progress=None,
        timeit=False,
    ):
        def set_paths(shapes, mapping):
            for obj in shapes["parts"]:
                if obj.get("parts") is None:
                    obj["ind"] = mapping[str(obj["id"])]["path"]
                else:
                    set_paths(obj, mapping)

        shapes = self.collect_shapes(
            loc=None,
            quality=quality,
            deviation=deviation,
            angular_tolerance=angular_tolerance,
            edge_accuracy=edge_accuracy,
            render_edges=render_edges,
            render_normals=render_normals,
            progress=progress,
            timeit=timeit,
        )
        set_paths(shapes, mapping)
        return shapes

    def to_state(self, parents=None):
        parents = parents or ()
        result = {}
        for i, obj in enumerate(self.objects):
            if isinstance(obj, _PartGroup):
                for k, v in obj.to_state((*parents, i)).items():
                    result[k] = v
            else:
                result[str(obj.id)] = {"path": (*parents, i), "state": obj.to_state()}
        return result

    def count_shapes(self):
        def c(pg):
            count = 0
            for p in pg.objects:
                if isinstance(p, _PartGroup):
                    count += c(p)
                else:
                    count += 1
            return count

        return c(self)

    @staticmethod
    def reset_id():
        global PART_ID
        PART_ID = 0

    def compounds(self):
        result = []
        for obj in self.objects:
            result += obj.compounds()
        return result

    def compound(self):
        return Compound._makeCompound(self.compounds())


def _combined_bb(shapes):
    def c_bb(shapes, bb):
        for shape in shapes["parts"]:
            if shape.get("parts") is None:
                if bb is None:
                    bb = BoundingBox(shape["bb"])
                else:
                    bb.update(shape["bb"])
            else:
                c_bb(shape, bb)
        return bb
    
    bb = c_bb(shapes, None)
    return bb


def _show(part_group, **kwargs):
    for k in kwargs:
        if get_default(k, "n/a") == "n/a":
            raise KeyError(f"Paramater {k} is not a valid argument for show()")

    if kwargs.get("cad_width") is not None and kwargs.get("cad_width") < 640:
        warn("cad_width has to be >= 640, setting to 640")
        kwargs["cad_width"] = 640

    if kwargs.get("height") is not None and kwargs.get("height") < 400:
        warn("height has to be >= 400, setting to 400")
        kwargs["height"] = 400

    if kwargs.get("tree_width") is not None and kwargs.get("tree_width") < 250:
        warn("tree_width has to be >= 250, setting to 250")
        kwargs["tree_width"] = 250

    # remove all tessellation and view parameters
    create_args, add_shape_args = split_args(kwargs)

    preset = lambda key, value: get_default(key) if value is None else value

    timeit = preset("timeit", kwargs.get("timeit"))

    with Timer(timeit, "", "overall"):

        with Timer(timeit, "", "setup display", 1):
            num_shapes = part_group.count_shapes()
            d = get_or_create_display(**create_args)
            d.init_progress(2 * num_shapes)

        with Timer(timeit, "", "tessellate", 1):

            mapping = part_group.to_state()
            shapes = part_group.collect_mapped_shapes(
                mapping,
                quality=preset("quality", kwargs.get("quality")),
                deviation=preset("deviation", kwargs.get("deviation")),
                angular_tolerance=preset("angular_tolerance", kwargs.get("angular_tolerance")),
                edge_accuracy=preset("edge_accuracy", kwargs.get("edge_accuracy")),
                render_edges=preset("render_edges", kwargs.get("render_edges")),
                render_normals=preset("render_normals", kwargs.get("render_normals")),
                progress=d.progress,
                timeit=timeit,
            )
            tree = part_group.to_nav_dict()

        with Timer(timeit, "", "show shapes", 1):
            d.add_shapes(shapes=shapes, mapping=mapping, tree=tree, bb=_combined_bb(shapes), **add_shape_args)

    d.info.version_msg(__version__)
    d.info.ready_msg(d.cq_view.grid.step)

    sidecar = has_sidecar()
    if sidecar is not None:
        print(f"Done, using side car '{sidecar.title()}'")

    return d



from array import array

import numpy as np

from OCP.gp import gp_Vec, gp_Pnt
from OCP.BRep import BRep_Tool
from OCP.BRepTools import BRepTools
from OCP.BRepGProp import BRepGProp_Face
from OCP.BRepMesh import BRepMesh_IncrementalMesh
from OCP.TopLoc import TopLoc_Location
from OCP.TopAbs import TopAbs_Orientation
from OCP.TopTools import TopTools_IndexedDataMapOfShapeListOfShape, TopTools_IndexedMapOfShape
from OCP.TopExp import TopExp, TopExp_Explorer
from OCP.TopAbs import TopAbs_EDGE, TopAbs_FACE, TopAbs_SOLID
from OCP.TopoDS import TopoDS
from OCP.BRepAdaptor import BRepAdaptor_Curve
from OCP.GCPnts import GCPnts_QuasiUniformDeflection

from jupyter_cadquery.utils import Timer
from jupyter_cadquery.ocp_utils import get_faces, bounding_box
from cadquery.occ_impl.shapes import Compound

# class RenderCache:
#     def __init__(self):
#         self.objects = {}
#         self.use_cache = True

#     def reset_cache(self):
#         self.objects = {}

#     def toggle_cache(self):
#         self.use_cache = not self.use_cache
#         print(f"Render cache turned {'ON' if self.use_cache else 'OFF'}")

#     def tessellate(
#         self,
#         compound,
#         quality=None,
#         angular_tolerance=None,
#         render_edges=True,
#         normals_len=0,
#         debug=False,
#     ):

#         hash = id(compound)  # use python id instead of compound.HashCode(HASH_CODE_MAX)
#         if self.objects.get(hash) is None:
#             tess = Tessellator()
#             tess.compute(
#                 compound,
#                 quality=quality,
#                 angular_tolerance=angular_tolerance,
#                 compute_edges=render_edges,
#                 normals_len=normals_len,
#                 debug=debug,
#             )
#             np_vertices = tess.get_vertices()
#             np_triangles = tess.get_triangles()
#             np_normals = tess.get_normals()
#             np_edges = tess.get_edges()

#             if np_normals.shape != np_vertices.shape:
#                 raise AssertionError("Wrong number of normals/shapes")

#             shape_geometry = BufferGeometry(
#                 attributes={
#                     "position": BufferAttribute(np_vertices),
#                     "index": BufferAttribute(np_triangles),
#                     "normal": BufferAttribute(np_normals),
#                 }
#             )
#             if debug:
#                 print(f"| | | (Caching {hash})")
#             self.objects[hash] = (shape_geometry, np_edges)
#         else:
#             if debug:
#                 print(f"| | | (Taking {hash} from cache)")
#         return self.objects[hash]


# RENDER_CACHE = RenderCache()
# reset_cache = RENDER_CACHE.reset_cache
# toggle_cache = RENDER_CACHE.toggle_cache
class Tessellator:
    def __init__(self):
        self.vertices = np.empty((0, 3), dtype="float32")
        self.triangles = np.empty((0,), dtype="uint32")
        self.normals = np.empty((0, 3), dtype="float32")
        self.normals = np.empty((0, 2, 3), dtype="float32")
        self.shape = None

    def number_solids(self, shape):
        count = 0
        e = TopExp_Explorer(shape, TopAbs_SOLID)
        while e.More():
            count += 1
            e.Next()
        return count

    def compute(
        self,
        shape,
        quality,
        angular_tolerance,
        tessellate=True,
        compute_edges=True,
        normals_len=0,
        debug=False,
    ):
        self.shape = shape
        self.normals_len = normals_len
        self.edges = []

        count = self.number_solids(shape)
        with Timer(debug, "", f"mesh incrementally {'(parallel)' if count > 1 else ''}", 3):
            # Remove previous mesh data
            BRepTools.Clean_s(shape)
            BRepMesh_IncrementalMesh(shape, quality, False, angular_tolerance, count > 1)

        if tessellate:
            with Timer(debug, "", "get nodes, triangles and normals", 3):
                self.tessellate()

        if compute_edges:
            with Timer(debug, "", "get edges", 3):
                self.compute_edges()

        # Remove mesh data again
        # BRepTools.Clean_s(shape)

    def tessellate(self):
        self.vertices = array("f")
        self.triangles = array("f")
        self.normals = array("f")

        # global buffers
        p_buf = gp_Pnt()
        n_buf = gp_Vec()
        loc_buf = TopLoc_Location()

        offset = -1

        # every line below is selected for performance. Do not introduce functions to "beautify" the code

        for face in get_faces(self.shape):
            if face.Orientation() == TopAbs_Orientation.TopAbs_REVERSED:
                i1, i2 = 2, 1
            else:
                i1, i2 = 1, 2

            internal = face.Orientation() == TopAbs_Orientation.TopAbs_INTERNAL

            poly = BRep_Tool.Triangulation_s(face, loc_buf)
            if poly is not None:
                Trsf = loc_buf.Transformation()

                # add vertices
                # [node.Transformed(Trsf).Coord() for node in poly.Nodes()] is 5-8 times slower!
                items = poly.Nodes()
                coords = [items.Value(i).Transformed(Trsf).Coord() for i in range(items.Lower(), items.Upper() + 1)]
                flat = []
                for coord in coords:
                    flat += coord
                self.vertices.extend(flat)

                # add triangles
                items = poly.Triangles()
                coords = [items.Value(i).Get() for i in range(items.Lower(), items.Upper() + 1)]
                flat = []
                for coord in coords:
                    flat += (coord[0] + offset, coord[i1] + offset, coord[i2] + offset)
                self.triangles.extend(flat)

                # add normals
                if poly.HasUVNodes():

                    def extract(uv0, uv1):
                        prop.Normal(uv0, uv1, p_buf, n_buf)
                        if n_buf.SquareMagnitude() > 0:
                            n_buf.Normalize()
                        return n_buf.Reverse().Coord() if internal else n_buf.Coord()

                    prop = BRepGProp_Face(face)
                    items = poly.UVNodes()

                    uvs = [items.Value(i).Coord() for i in range(items.Lower(), items.Upper() + 1)]
                    flat = []
                    for uv1, uv2 in uvs:
                        flat += extract(uv1, uv2)
                    self.normals.extend(flat)

                offset += poly.NbNodes()

    def compute_edges(self):
        edge_map = TopTools_IndexedMapOfShape()
        face_map = TopTools_IndexedDataMapOfShapeListOfShape()

        TopExp.MapShapes_s(self.shape, TopAbs_EDGE, edge_map)
        TopExp.MapShapesAndAncestors_s(self.shape, TopAbs_EDGE, TopAbs_FACE, face_map)

        for i in range(1, edge_map.Extent() + 1):
            edge = TopoDS.Edge_s(edge_map.FindKey(i))

            face_list = face_map.FindFromKey(edge)
            if face_list.Extent() == 0:
                # print("no faces")
                continue

            loc = TopLoc_Location()
            poly = BRep_Tool.Polygon3D_s(edge, loc)

            if poly is not None:
                # print("Polygon3D successful")
                nodes = poly.Nodes()
                transf = loc.Transformation()
                v1 = None
                for j in range(1, poly.NbNodes() + 1):
                    v2 = nodes.Value(j)
                    v2.Transform(transf)
                    v2 = v2.Coord()
                    if v1 is not None:
                        self.edges.append((v1, v2))
                    v1 = v2
            else:
                face = TopoDS.Face_s(face_list.First())
                triang = BRep_Tool.Triangulation_s(face, loc)
                poly = BRep_Tool.PolygonOnTriangulation_s(edge, triang, loc)
                if poly is None:
                    continue

                indices = poly.Nodes()
                nodes = triang.Nodes()
                transf = loc.Transformation()
                v1 = None
                for j in range(indices.Lower(), indices.Upper() + 1):
                    v2 = nodes.Value(indices.Value(j))
                    v2.Transform(transf)
                    v2 = v2.Coord()
                    if v1 is not None:
                        self.edges.append((v1, v2))
                    v1 = v2

    def get_vertices(self):
        return np.asarray(self.vertices, dtype=np.float32).reshape(-1, 3)

    def get_triangles(self):
        return np.asarray(self.triangles, dtype=np.uint32)

    def get_normals(self):
        return np.asarray(self.normals, dtype=np.float32).reshape(-1, 3)

    def get_edges(self):
        normal_edges = []
        if self.normals_len > 0:
            vertices = self.get_vertices()
            normals = self.get_normals()
            normal_edges = np.column_stack((vertices, vertices + (normals * self.normals_len))).reshape((-1, 2, 3))

        return (np.asarray(self.edges, dtype=np.float32), normal_edges)


def compute_quality(bb, deviation=0.1):
    return (bb.xsize + bb.ysize + bb.zsize) / 300 * deviation


def tessellate(
    shapes,
    quality: float,
    angular_tolerance: float,
    tessellate=True,
    compute_edges=True,
    normals_len=0,
    debug=False,
):
    compound = Compound._makeCompound(shapes) if len(shapes) > 1 else shapes[0]
    tess = Tessellator()
    tess.compute(compound, quality, angular_tolerance, tessellate, compute_edges, normals_len, debug)
    return {
        "vertices": tess.get_vertices(),
        "triangles": tess.get_triangles(),
        "normals": tess.get_normals(),
        "edges": tess.get_edges(),
    }


def discretize_edge(edge, deflection=0.1):
    curve_adaptator = BRepAdaptor_Curve(edge)

    discretizer = GCPnts_QuasiUniformDeflection()
    discretizer.Initialize(
        curve_adaptator, deflection, curve_adaptator.FirstParameter(), curve_adaptator.LastParameter()
    )

    if not discretizer.IsDone():
        raise AssertionError("Discretizer not done.")

    points = [curve_adaptator.Value(discretizer.Parameter(i)).Coord() for i in range(1, discretizer.NbPoints() + 1)]

    # return tuples representing the single lines of the egde
    edges = []
    for i in range(len(points) - 1):
        edges.append((points[i], points[i + 1]))

    return np.asarray(edges, dtype=np.float32)


#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

try:
    import cqparts
    from cqparts.display.material import COLOR
    from cqparts.utils.geometry import CoordSystem

    has_cqparts = True
except:
    has_cqparts = False

import jupyter_cadquery as jcq


def is_cqparts_assembly(cad_obj):
    return has_cqparts and has_cqparts and isinstance(cad_obj, cqparts.Assembly)


def is_cqparts_part(cad_obj):
    return has_cqparts and has_cqparts and isinstance(cad_obj, cqparts.Part)


def is_cqparts(cad_obj):
    return is_cqparts_assembly(cad_obj) or is_cqparts_part(cad_obj)


def convert_cqparts(cad_obj, name="root", default_color=None, replay=False):
    cad_obj.world_coords = CoordSystem()
    return _convert_cqparts(cad_obj, name, default_color, replay)


def _convert_cqparts(cad_obj, name, default_color, replay):
    if default_color is None:
        default_color = (255, 255, 0)

    if isinstance(cad_obj, cqparts.Assembly):
        parts = []
        for k, v in cad_obj._components.items():
            parts.append(_convert_cqparts(v, k, default_color, replay))
        return jcq.cadquery.PartGroup(parts, name)

    elif isinstance(cad_obj, cqparts.Part):
        if replay:
            return cad_obj.world_obj
        else:
            color = cad_obj._render.color
            if isinstance(color, (list, tuple)):
                pass
            elif isinstance(color, str):
                color = COLOR.get(color)
                if color is None:
                    color = default_color
            else:
                color = default_color
            color = tuple((c / 255.0 for c in color))
            return jcq.cadquery.Part(cad_obj.world_obj, name, color)
    else:
        print("No cqparts object")
        return None


def get_parts(cad_obj, name="root"):
    parts = {}
    if isinstance(cad_obj, cqparts.Assembly):
        for k, v in cad_obj._components.items():
            parts[k] = get_parts(v, k)
        return parts
    elif isinstance(cad_obj, cqparts.Part):
        return cad_obj.world_obj



import itertools
import numpy as np

from OCP.Bnd import Bnd_Box
from OCP.BRep import BRep_Tool
from OCP.BRepBndLib import BRepBndLib
from OCP.BRepMesh import BRepMesh_IncrementalMesh
from OCP.BRepTools import BRepTools


from OCP.TopAbs import (
    TopAbs_EDGE,
    TopAbs_FACE,
)
from OCP.TopoDS import TopoDS_Compound
from OCP.TopAbs import TopAbs_FACE
from OCP.TopExp import TopExp_Explorer

from OCP.StlAPI import StlAPI_Writer

from cadquery import Compound, Location
from cadquery.occ_impl.shapes import downcast
from .utils import distance


HASH_CODE_MAX = 2147483647


class BoundingBox(object):
    def __init__(self, obj=None, optimal=False, tol=1e-5):
        self.optimal = optimal
        self.tol = tol
        if obj is None:
            self.xmin = self.xmax = self.ymin = self.ymax = self.zmin = self.zmax = 0
        elif isinstance(obj, BoundingBox):
            self.xmin = obj.xmin
            self.xmax = obj.xmax
            self.ymin = obj.ymin
            self.ymax = obj.ymax
            self.zmin = obj.zmin
            self.zmax = obj.zmax
        elif isinstance(obj, dict):
            self.xmin = obj["xmin"]
            self.xmax = obj["xmax"]
            self.ymin = obj["ymin"]
            self.ymax = obj["ymax"]
            self.zmin = obj["zmin"]
            self.zmax = obj["zmax"]    
        else:
            bbox = self._bounding_box(obj, tol)
            self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax = bbox
        
        self._calc()

    def _bounding_box(self, obj, tol=1e-5):
        bbox = Bnd_Box()
        if self.optimal:
            BRepTools.Clean_s(obj)
            BRepBndLib.AddOptimal_s(obj, bbox)
        else:
            BRepBndLib.Add_s(obj, bbox)
        values = bbox.Get()
        return (values[0], values[3], values[1], values[4], values[2], values[5])

    def _calc(self):
        self.xsize = self.xmax - self.xmin
        self.ysize = self.ymax - self.ymin
        self.zsize = self.zmax - self.zmin
        self.center = (
            self.xmin + self.xsize / 2.0,
            self.ymin + self.ysize / 2.0,
            self.zmin + self.zsize / 2.0,
        )
        self.max = max([abs(x) for x in (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax)])

    def is_empty(self, eps=0.01):
        return (
            (abs(self.xmax - self.xmin) < 0.01)
            and (abs(self.ymax - self.ymin) < 0.01)
            and (abs(self.zmax - self.zmin) < 0.01)
        )

    def max_dist_from_center(self):
        return max(
            [
                distance(self.center, v)
                for v in itertools.product((self.xmin, self.xmax), (self.ymin, self.ymax), (self.zmin, self.zmax))
            ]
        )

    def max_dist_from_origin(self):
        return max(
            [
                np.linalg.norm(v)
                for v in itertools.product((self.xmin, self.xmax), (self.ymin, self.ymax), (self.zmin, self.zmax))
            ]
        )

    def update(self, bb, minimize=False):
        lower, upper = (max, min) if minimize else (min, max)

        if isinstance(bb, BoundingBox):
            self.xmin = lower(bb.xmin, self.xmin)
            self.xmax = upper(bb.xmax, self.xmax)
            self.ymin = lower(bb.ymin, self.ymin)
            self.ymax = upper(bb.ymax, self.ymax)
            self.zmin = lower(bb.zmin, self.zmin)
            self.zmax = upper(bb.zmax, self.zmax)
        elif isinstance(bb, dict):
            self.xmin = lower(bb["xmin"], self.xmin)
            self.xmax = upper(bb["xmax"], self.xmax)
            self.ymin = lower(bb["ymin"], self.ymin)
            self.ymax = upper(bb["ymax"], self.ymax)
            self.zmin = lower(bb["zmin"], self.zmin)
            self.zmax = upper(bb["zmax"], self.zmax)
        else:
            raise "Wrong bounding box param"

        self._calc()

    def to_dict(self):
        return {
            "xmin": self.xmin,
            "xmax": self.xmax,
            "ymin": self.ymin,
            "ymax": self.ymax,
            "zmin": self.zmin,
            "zmax": self.zmax,
        }

    def __repr__(self):
        return "{xmin:%.2f, xmax:%.2f, ymin:%.2f, ymax:%.2f, zmin:%.2f, zmax:%.2f}" % (
            self.xmin,
            self.xmax,
            self.ymin,
            self.ymax,
            self.zmin,
            self.zmax,
        )


def bounding_box(objs, loc=None, optimal=False):
    if isinstance(objs, (list, tuple)):
        compound = Compound._makeCompound(objs)
    else:
        compound = objs

    return BoundingBox(compound if loc is None else compound.Moved(loc.wrapped), optimal=optimal)


# Export STL


def write_stl_file(compound, filename, tolerance=None, angular_tolerance=None):

    # Remove previous mesh data
    BRepTools.Clean_s(compound)

    mesh = BRepMesh_IncrementalMesh(compound, tolerance, True, angular_tolerance)
    mesh.Perform()

    writer = StlAPI_Writer()

    result = writer.Write(compound, filename)

    # Remove the mesh data again
    BRepTools.Clean_s(compound)
    return result


# OCP types and accessors


def is_compound(topods_shape):
    return isinstance(topods_shape, TopoDS_Compound)


def _get_topo(shape, topo):
    explorer = TopExp_Explorer(shape, topo)
    hashes = {}
    while explorer.More():
        item = explorer.Current()
        hash = item.HashCode(HASH_CODE_MAX)
        if hashes.get(hash) is None:
            hashes[hash] = True
            yield downcast(item)
        explorer.Next()


def get_faces(shape):
    return _get_topo(shape, TopAbs_FACE)


def get_edges(shape):
    return _get_topo(shape, TopAbs_EDGE)


def get_point(vertex):
    p = BRep_Tool.Pnt_s(vertex)
    return (p.X(), p.Y(), p.Z())


def get_rgb(color):
    if color is None:
        return (176, 176, 176)
    rgb = color.wrapped.GetRGB()
    return (int(255 * rgb.Red()), int(255 * rgb.Green()), int(255 * rgb.Blue()))


def loc_to_tq(loc):
    T = loc.wrapped.Transformation()
    t = T.Transforms()
    q = T.GetRotation()
    return (t, (q.X(), q.Y(), q.Z(), q.W()))


def __location__repr__(self):
    t, r = loc_to_tq(self)
    return f"(t={t}, q=({r[0]}, {r[1]}, {r[2]}, {r[3]}))"


Location.__repr__ = __location__repr__  # type: ignore



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
from os.path import join, dirname
import pickle
from uuid import uuid4
from IPython.display import display as ipy_display

from ipywidgets import Label, Checkbox, Layout, HBox, VBox, Box, FloatSlider, Tab, HTML, Box, Output

from jupyter_cadquery_widgets.widgets import ImageButton, TreeView, UNSELECTED, SELECTED, MIXED, EMPTY
from .cad_view import CadqueryView
from .utils import Timer, Progress, px
from ._version import __version__
from .defaults import set_defaults, get_default, split_args
from .logo import LOGO_DATA
from .style import set_css

DISPLAY = None
SIDECAR = None


def has_sidecar():
    if SIDECAR is not None:
        return SIDECAR.title
    else:
        return None


def close_sidecar():
    global DISPLAY, SIDECAR

    if SIDECAR is not None:
        SIDECAR.close()

    DISPLAY, SIDECAR = None, None


def set_sidecar(title, init=False):
    global SIDECAR
    try:
        from sidecar import Sidecar

        _has_sidecar = True
    except:
        print("Warning: module sidecar not installed")
        _has_sidecar = False

    if _has_sidecar:
        close_sidecar()
        SIDECAR = Sidecar(title=title)
        set_defaults(display="sidecar")

        if init:
            d = get_or_create_display(init=True)


def reset_sidecar(init=True):
    if SIDECAR is None:
        title = "CadQuery"
    else:
        title = SIDECAR.title

    close_sidecar()
    set_sidecar(title, init)


def get_or_create_display(init=False, **kwargs):
    global DISPLAY

    def resize():
        t = kwargs.get("tree_width")
        w = kwargs.get("cad_width")
        h = kwargs.get("height")
        if w is not None or h is not None:
            DISPLAY.set_size(t, w, h)

    if kwargs.get("display", get_default("display")) != "sidecar" or SIDECAR is None:
        d = CadqueryDisplay()
        widget = d.create(**kwargs)
        ipy_display(widget)
        set_css(get_default("theme"))
        return d

    if DISPLAY is None:
        DISPLAY = CadqueryDisplay()
        widget = DISPLAY.create(**kwargs)
        resize()
        SIDECAR.clear_output(True)
        with SIDECAR:
            ipy_display(widget)

        data = pickle.loads(base64.b64decode(LOGO_DATA))
        mesh_data = data["data"]
        config = data["config"]
        DISPLAY.init_progress(data.get("count", 1))
        create_args, add_shape_args = split_args(config)
        DISPLAY._update_settings(**create_args)
        DISPLAY.add_shapes(**mesh_data, **add_shape_args)
        DISPLAY.info.ready_msg(DISPLAY.cq_view.grid.step)
        DISPLAY.splash = True
        set_css(get_default("theme"), True)

    else:
        # Use the existing Cad Display, so set the defaults and parameters again
        DISPLAY._update_settings(**kwargs)
        resize()
        set_css(get_default("theme"))

    # Change latest settings
    # DISPLAY._update_settings(**kwargs)
    DISPLAY.cq_view.timeit = kwargs.get("timeit", get_default("timeit"))

    return DISPLAY


class Info(object):
    def __init__(self, width=230, height=300):
        self.html = HTML(
            value="",
            layout=Layout(
                width=(px(width)),
                height=(px(height)),
                border="solid 1px #ddd",
                overflow="scroll",
            ),
        )
        self.width = width
        self.height = height
        self.number = 0
        self.chunks = []

    def clear(self):
        self.html.value = ""
        self.number = 0
        self.chunks = []

    def add_text(self, msg):
        self.add_html('<pre style="white-space: nowrap;">%s</pre>' % msg)

    def add_html(self, html):
        self.chunks.insert(0, (self.number, html))
        self.number += 1
        self.render()

    def render(self):
        html = '<table style="display: block; overflow-x: visible; white-space: nowrap;">'

        for n, chunk in self.chunks:
            html += '<tr style="vertical-align: text-top;">'
            html += '<td><pre style="color: #aaa; white-space: nowrap">[%2d]</pre></td>' % n
            html += "<td>%s</td>" % chunk
            html += "</tr>"
        html += "</table>"

        self.html.value = html

    def version_msg(self, version):
        self.add_html(
            f"""
        <b>Versions</b>
        <table>
            <tr class="small_table"><td>CadQuery:</td>        <td>{version}</td> </tr>
            <tr class="small_table"><td>Jupyter CadQuery:</td><td>{__version__}</td> </tr>
        </table>
        """
        )

    def ready_msg(self, tick_size):
        html = (
            """
        <b>Rendering done</b>
        <table>
            <tr class="small_table" >                      <td>Tick size</td>  <td>%s mm</td> </tr>
            <tr class="small_table" style="color: red;">   <td>X-Axis</td>     <td>Red</td>    </tr>
            <tr class="small_table" style="color: green;"> <td>Y-Axis</td>     <td>Green</td>  </tr>
            <tr class="small_table" style="color: blue;">  <td>Z-Axis</td>     <td>Blue</td>   </tr>
        </table>
        """
            % tick_size
        )
        self.add_html(html)

    def bb_info(self, name, bb):
        html = (
            """
        <b> Object: '%s':</b>
        <table>
        """
            % name
        )
        html += '<tr class="small_table"><th></th><th>min</th><th>max</th><th>center</th></tr>'

        for t, a, c in zip(("x", "y", "z"), bb[:3], bb[3]):
            html += """<tr class="small_table">
                <th>%s</th><td align='right'>%5.2f</td><td align='right'>%5.2f</td><td align='right'>%5.2f</td>
            </tr>
            """ % (
                t,
                a[0],
                a[1],
                c,
            )
        html += "</table>"
        self.add_html(html)


class Clipping(object):
    def __init__(self, image_path, out, cq_view, width):
        self.image_path = image_path
        self.cq_view = cq_view
        self.out = out
        self.width = width
        self.sliders = []
        self.normals = []
        self.labels = []

    def handler(self, b):
        i = int(b.type)
        self.cq_view.set_plane(i)
        self.labels[i].value = "N=(%5.2f, %5.2f, %5.2f)" % tuple(self.cq_view.direction())

    def set_width(self, width):
        self.width = width
        for slider in self.sliders:
            slider.layout.width = px(width - 20)

    def slider(self, value, min, max, step, description):
        label = Label(description)
        self.labels.append(label)
        ind = len(self.normals)
        button = ImageButton(
            width=36,
            height=28,
            image_path=join(self.image_path, "plane.png"),
            tooltip="Set clipping plane",
            type=str(ind),
            layout=Layout(margin="0px 10px 0px 0px"),
        )
        button.on_click(self.handler)
        button.add_class("view_button")

        slider = FloatSlider(
            value=value,
            min=min,
            max=max,
            step=step,
            description="",
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format=".2f",
            layout=Layout(width=px(self.width - 20)),
        )

        slider.observe(self.cq_view.clip(ind), "value")
        return [HBox([button, label]), slider]

    def add_slider(self, value, v_min, v_max, step, normal):
        self.sliders += self.slider(value, v_min, v_max, step, "N=(%5.2f, %5.2f, %5.2f)" % normal)
        self.normals.append(normal)

    def create(self):
        return VBox(self.sliders, layout=Layout(overflow="hidden"))


class CadqueryDisplay(object):

    types = [
        "reset",
        "fit",
        "isometric",
        "right",
        "front",
        "left",
        "rear",
        "top",
        "bottom",
    ]
    directions = {
        "left": (1, 0, 0),
        "right": (-1, 0, 0),
        "front": (0, 1, 0),
        "rear": (0, -1, 0),
        "top": (0, 0, 1),
        "bottom": (0, 0, -1),
        "isometric": (1, 1, 1),
    }

    def __init__(self):
        super().__init__()
        self.info = None
        self.cq_view = None
        self.assembly = None

        self.image_path = join(dirname(__file__), "icons", get_default("theme"))

        self.image_paths = [
            {
                UNSELECTED: join(self.image_path, "no_shape.png"),
                SELECTED: join(self.image_path, "shape.png"),
                MIXED: join(self.image_path, "mix_shape.png"),
                EMPTY: join(self.image_path, "empty_shape.png"),
            },
            {
                UNSELECTED: join(self.image_path, "no_mesh.png"),
                SELECTED: join(self.image_path, "mesh.png"),
                MIXED: join(self.image_path, "mix_mesh.png"),
                EMPTY: join(self.image_path, "empty_mesh.png"),
            },
        ]

        self._display = "cell"
        self._tools = True
        self.id = uuid4().hex[:10]
        self.clean = True
        self.splash = False
        self.tree_clipping = None

    def _dump_config(self):
        print("\nCadDisplay:")
        config = {
            k: v
            for k, v in self.__dict__.items()
            if not k
            in [
                "cq_view",
                "output",
                "info",
                "clipping",
                "tree_clipping",
                "image_paths",
                "image_path",
                "view_controls",
                "check_controls",
            ]
        }
        for k, v in config.items():
            print(f"- {k:30s}: {v}")

        print("\nCadView:")
        config = {
            k: v
            for k, v in self.cq_view.__dict__.items()
            if not k
            in [
                "pickable_objects",
                "scene",
                "controller",
                "renderer",
                "key_lights",
                "picker",
                "shapes",
                "camera",
                "info",
                "axes",
                "grid",
                "cq_renderer",
            ]
        }
        for k, v in config.items():
            print(f"- {k:30s}: {v}")

        print("\nCamera:")
        config = {
            k: v
            for k, v in self.cq_view.camera.__dict__["_trait_values"].items()
            if not k
            in [
                "keys",
                "matrix",
                "matrixWorldInverse",
                "modelViewMatrix",
                "normalMatrix",
                "matrixWorld",
                "projectionMatrix",
                "comm",
            ]
            and not k.startswith("_")
        }
        for k, v in config.items():
            print(f"- {k:30s}: {v}")

    # Buttons

    def create_button(self, image_name, handler, tooltip):
        button = ImageButton(
            width=36,
            height=28,
            image_path=join(self.image_path, f"{image_name}.png"),
            tooltip=tooltip,
            type=image_name,
        )
        button.on_click(handler)
        button.add_class("view_button")
        return button

    def create_checkbox(self, kind, description, value, handler):
        checkbox = Checkbox(value=value, description=description, indent=False)
        checkbox.observe(handler, "value")
        checkbox.add_class("view_%s" % kind)
        return checkbox

    # UI Handler

    def change_view(self, typ, directions):
        def reset(b):
            self.cq_view._reset_camera()

        def refit(b):
            self.cq_view.camera.zoom = self.cq_view.camera_initial_zoom
            self.cq_view._update()

        def change(b):
            self.cq_view.camera.position = self.cq_view._add(
                self.cq_view.bb.center, self.cq_view._scale(directions[typ])
            )
            self.cq_view._update()

        if typ == "fit":
            return refit
        elif typ == "reset":
            return reset
        else:
            return change

    def bool_or_new(self, val):
        return val if isinstance(val, bool) else val["new"]

    def toggle_axes(self, change):
        self.axes = self.bool_or_new(change)
        self.cq_view.set_axes_visibility(self.axes)

    def toggle_grid(self, change):
        self.grid = self.bool_or_new(change)
        self.cq_view.set_grid_visibility(self.grid)

    def toggle_axes0(self, change):
        self.axes0 = self.bool_or_new(change)
        self.cq_view.set_axes_center(self.axes0)

    def toggle_ortho(self, change):
        self.ortho = self.bool_or_new(change)
        self.cq_view.toggle_ortho(self.ortho)

    def toggle_transparent(self, change):
        self.transparent = self.bool_or_new(change)
        self.cq_view.set_transparent(self.transparent)

    def toggle_black_edges(self, change):
        self.black_edges = self.bool_or_new(change)
        self.cq_view.set_black_edges(self.black_edges)

    def toggle_clipping(self, change):
        if change["name"] == "selected_index":
            self.cq_view.set_clipping(change["new"])

    def init_progress(self, num_shapes):
        self.progress.progress.value = 0
        self.progress.reset(num_shapes)

    def _set_checkboxes(self):
        self.checkbox_axes.value = self.axes
        self.checkbox_grid.value = self.grid
        self.checkbox_axes0.value = self.axes0
        self.checkbox_ortho.value = self.ortho
        self.checkbox_transparent.value = self.transparent
        self.checkbox_black_edges.value = self.black_edges

    def _update_settings(self, **kwargs):
        preset = lambda key, value: get_default(key) if value is None else value

        self.height = preset("height", kwargs.get("height"))
        self.tree_width = preset("tree_width", kwargs.get("tree_width"))
        self.cad_width = preset("cad_width", kwargs.get("cad_width"))
        self.bb_factor = preset("bb_factor", kwargs.get("bb_factor"))
        self.axes = preset("axes", kwargs.get("axes"))
        self.axes0 = preset("axes0", kwargs.get("axes0"))
        self.grid = preset("grid", kwargs.get("grid"))
        self.ortho = preset("ortho", kwargs.get("ortho"))
        self.transparent = preset("transparent", kwargs.get("transparent"))
        self.black_edges = preset("black_edges", kwargs.get("black_edges"))
        self.mac_scrollbar = preset("mac_scrollbar", kwargs.get("mac_scrollbar"))
        self.timeit = preset("timeit", kwargs.get("timeit"))
        self._display = preset("display", kwargs.get("display"))
        self._tools = preset("tools", kwargs.get("tools"))

    def _info_height(self, height):
        return int(height * 0.4) - 20 + 2

    def _tree_clipping_height(self, height):
        return int(height * 0.6) + 17

    def _tree_height(self, height):
        return int(height * 0.6) - 30

    def set_size(self, tree_width, width, height):
        if width is not None:
            # adapt renderer
            self.cad_width = width
            self.cq_view.camera.width = width
            self.cq_view.renderer.width = width

        if height is not None:
            # adapt renderer
            self.height = height
            self.cq_view.camera.height = height
            self.cq_view.renderer.height = height
            # adapt info box
            info_height = self._info_height(height)
            self.info.height = info_height
            self.info.html.layout.height = px(info_height - 6)
            self.output.layout.height = px(info_height)
            # adapt tree and clipping
            tree_clipping_height = self._tree_clipping_height(height)
            self.tree_clipping.layout.height = px(tree_clipping_height)

            tree_height = self._tree_height(height)
            self.tree_view.layout.height = px(tree_height)

        if tree_width is not None:
            # adapt tree and clipping
            self.tree_clipping.layout.width = px(tree_width)
            self.tree_view.layout.width = px(tree_width)
            self.clipping.set_width(tree_width)
            # adapt info box
            self.output.layout.width = px(tree_width)
            self.info.html.layout.width = px(tree_width)
            # adapt progress bar
            self.progress.progress.layout.width = px(tree_width)

    def create(
        self,
        height=None,
        bb_factor=None,
        tree_width=None,
        cad_width=None,
        axes=None,
        axes0=None,
        grid=None,
        ortho=None,
        transparent=None,
        mac_scrollbar=None,
        display=None,
        tools=None,
        timeit=None,
    ):
        self._update_settings(
            height=height,
            bb_factor=bb_factor,
            tree_width=tree_width,
            cad_width=cad_width,
            axes=axes,
            axes0=axes0,
            grid=grid,
            ortho=ortho,
            transparent=transparent,
            mac_scrollbar=mac_scrollbar,
            display=display,
            tools=tools,
            timeit=timeit,
        )

        # Output widget
        output_height = self._info_height(self.height)
        self.info = Info(self.tree_width, output_height - 6)
        self.info.html.add_class("scroll-area")
        if self.mac_scrollbar:
            self.info.html.add_class("mac-scrollbar")

        ## Threejs rendering of Cadquery objects
        self.cq_view = CadqueryView(
            width=self.cad_width,
            height=self.height,
            info=self.info,
            timeit=self.timeit,
        )

        renderer = self.cq_view.create()
        renderer.add_class("view_renderer")
        renderer.add_class(f"view_renderer_{self.id}")

        # Prepare the CAD view tools
        # Output area

        self.output = Box([self.info.html])
        self.output.layout = Layout(
            height=px(output_height),
            width=px(self.tree_width),
            overflow_y="hidden",
            overflow_x="hidden",
        )
        self.output.add_class("view_output")
        # TODO
        # if get_default("theme") == "dark":
        #     self.output.add_class("p-Collapse-contents")

        # Clipping tool
        self.clipping = Clipping(self.image_path, self.output, self.cq_view, self.tree_width)
        for normal in ((-1.0, 0.0, 0.0), (0.0, -1.0, 0.0), (0.0, 0.0, -1.0)):
            self.clipping.add_slider(1, -1, 1, 0.01, normal)

        # Empty dummy Tree View
        self.tree_view = Output()
        self.progress = Progress(3, self.tree_width)

        # Tab widget with Tree View and Clipping tools
        self.tree_clipping = Tab(
            layout=Layout(height=px(self._tree_clipping_height(self.height)), width=px(self.tree_width))
        )
        self.tree_clipping.children = [self.tree_view, self.clipping.create()]
        for i, c in enumerate(["Tree", "Clipping"]):
            self.tree_clipping.set_title(i, c)
        self.tree_clipping.observe(self.toggle_clipping)
        self.tree_clipping.add_class("tab-content-no-padding")

        # Check controls to switch orto, grid and axis
        self.checkbox_axes = self.create_checkbox("axes", "Axes", self.axes, self.toggle_axes)
        self.checkbox_grid = self.create_checkbox("grid", "Grid", self.grid, self.toggle_grid)
        self.checkbox_axes0 = self.create_checkbox("zero", "@ 0", self.axes0, self.toggle_axes0)
        self.checkbox_ortho = self.create_checkbox("ortho", "Ortho", self.ortho, self.toggle_ortho)
        self.checkbox_transparent = self.create_checkbox(
            "transparent",
            "Transparency",
            self.transparent,
            self.toggle_transparent,
        )
        self.checkbox_black_edges = self.create_checkbox("black_edges", "Black Edges", False, self.toggle_black_edges)
        self.check_controls = [
            self.checkbox_axes,
            self.checkbox_grid,
            self.checkbox_axes0,
            self.checkbox_ortho,
            self.checkbox_transparent,
            self.checkbox_black_edges,
        ]
        self.check_controls[-2].add_class("indent")

        # Buttons to switch camera position
        self.view_controls = []
        for typ in CadqueryDisplay.types:
            if typ == "refit":
                tooltip = "Fit view"
            elif typ == "reset":
                tooltip = "Reset view"
            else:
                tooltip = "Change view to %s" % typ
            button = self.create_button(typ, self.change_view(typ, CadqueryDisplay.directions), tooltip)
            self.view_controls.append(button)

        # only show pure renderer
        if self._tools == False:
            return renderer
        else:
            return HBox(
                [
                    VBox([HBox(self.check_controls[:-2]), self.tree_clipping, self.progress.progress, self.output]),
                    VBox([HBox(self.view_controls + self.check_controls[-2:]), renderer]),
                ]
            )

    def add_shapes(
        self,
        shapes,
        mapping,
        tree,
        bb,
        ticks=None,
        reset_camera=True,
        bb_factor=None,
        ambient_intensity=None,
        direct_intensity=None,
        default_edgecolor=None,
        position=None,
        rotation=None,
        zoom=None,
    ):
        self.clear()
        self.states = {k: v["state"] for k, v in mapping.items()}
        self.paths = {k: v["path"] for k, v in mapping.items()}

        self.tree_view = Output()
        self.tree_clipping.children = [self.tree_view, self.tree_clipping.children[1]]

        # Force reset of camera to inhereit splash settings for first object
        if self.splash:
            reset_camera = True
            self.splash = False

        with Timer(self.timeit, "", "mesh shapes", 2):
            self.cq_view.add_shapes(
                shapes,
                bb,
                ticks,
                self.progress,
                reset_camera=reset_camera,
                bb_factor=bb_factor,
                ambient_intensity=ambient_intensity,
                direct_intensity=direct_intensity,
                default_edgecolor=default_edgecolor,
                position=position,
                rotation=rotation,
                zoom=zoom,
            )

        with Timer(self.timeit, "", "configure display", 2):

            def set_slider(i, s_min, s_max):
                s_min = -0.02 if abs(s_min) < 1e-4 else s_min * self.bb_factor
                s_max = 0.02 if abs(s_max) < 1e-4 else s_max * self.bb_factor
                self.clipping.sliders[i].max = 2 ** 31  #  first increase max to avoid traitlet error that min > max
                self.clipping.sliders[i].min = s_min  # set min which now is always < max
                self.clipping.sliders[i].max = s_max  # correct max
                self.clipping.sliders[i].value = s_max

            bb = self.cq_view.bb
            set_slider(1, bb.xmin, bb.xmax)
            set_slider(3, bb.ymin, bb.ymax)
            set_slider(5, bb.zmin, bb.zmax)

            # Tree widget to change visibility
            self.tree_view = TreeView(
                image_paths=self.image_paths,
                tree=tree,
                state=self.states,
                layout=Layout(height=px(self._tree_height(self.height)), width=px(self.tree_width - 20)),
            )
            self.tree_view.add_class("view_tree")
            self.tree_view.add_class("scroll-area")
            if self.mac_scrollbar:
                self.tree_view.add_class("mac-scrollbar")

            self.tree_view.observe(self.cq_view.change_visibility(self.paths), "state")
            self.tree_clipping.children = [self.tree_view, self.tree_clipping.children[1]]

            # Set initial state

            for obj, vals in self.states.items():
                for i, val in enumerate(vals):
                    self.cq_view.set_visibility(self.paths[obj], i, val)

            self._set_checkboxes()
            self.toggle_axes(self.axes)
            self.toggle_axes0(self.axes0)
            self.toggle_grid(self.grid)
            self.toggle_transparent(self.transparent)
            self.toggle_black_edges(self.black_edges)
            self.toggle_ortho(self.ortho)

            self.clean = False

    def clear(self):
        if not self.clean:
            self.cq_view.clear()
            self.info.clear()

            # clear tree
            self.tree_clipping.children = [Output(), self.tree_clipping.children[1]]

            self.clean = True

    def display(self, widget):
        if self._display == "cell" or SIDECAR is None:
            ipy_display(widget)
        else:
            SIDECAR.clear_output(True)
            with SIDECAR:
                ipy_display(widget)

    @property
    def root_group(self):
        return self.cq_view.root_group
